# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


"""Fillmore is a Python library for scrubbing Sentry events."""

# x.y.z or x.y.z.dev0
__version__ = "0.1.1"


SCRUBBER_MODULE_NAME = __name__
